' ------------------------------------------------------------------------------
' ILvlif 1.0 for Windows
' Copyright (c) 2009  J-F Garnier
'
' This program is free software; you can redistribute it and/or
' modify it under the terms of the GNU General Public License
' as published by the Free Software Foundation; either version 2
' of the License, or (at your option) any later version.
' 
' This program is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
' GNU General Public License for more details.
' 
' You should have received a copy of the GNU General Public License
' along with this program; if not, write to the Free Software
' Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
'
' ----------------------------------------------------------------------------------
'
' MainForm.vb  Main module, GUI, PIL-Box management
'
' Nov  2009: Version 1.0 
' ---------------------------------------------------------------------------------

Imports System.IO

' --------------------------------
' MainForm
' Main application form for ILvlif
' --------------------------------
Public Class MainForm

    Dim lasth As Integer = 0      ' store the last sent/received high byte on serial
    Dim frame_timeout As Boolean

   ' PIL-Box special init command frames:
    Const CON = &H496     ' CON
    Const COFF = &H497    ' COFF
    Const TDIS = &H494    ' TDIS

    Const USE8BITS = True ' use 8-bit format with PILBox

    Const CRLF = Chr(13) & Chr(10)  ' convenient string

    ' --------------------------------
    ' DisplayStr()
    ' display a string in the display box
    ' --------------------------------
    Public Sub DisplayStr(ByVal ch As String)
        TextBox_display.AppendText(ch)
    End Sub

    ' --------------------------------
    ' ClrDisplay()
    ' clear the display box
    ' --------------------------------
    Public Sub ClrDisplay()
        TextBox_display.Clear()
    End Sub

    ' --------------------------------
    ' DisplayMnemo()
    ' display a string in the scope box
    ' (if the scope is enabled)
    ' --------------------------------
    Public Sub DisplayMnemo(ByVal ch As String)
        Static cpt As Integer = 0
        If CheckBox_scope.Checked Then
            TextBox_scope.AppendText(ch)
            cpt = cpt + 1
            If cpt = 3 Then
                cpt = 0
                TextBox_scope.AppendText(CRLF)
            End If
        End If
    End Sub




    ' --------------------------------
    ' START button 
    ' --------------------------------
    Private Sub ButtonStart_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles ButtonStart.Click

        ClrDisplay()

        SerialPort1.PortName = TextBox_ComPort.Text
        If RadioButton_highspeed.Checked Then SerialPort1.BaudRate = 115200 Else SerialPort1.BaudRate = 9600

        Try
            SerialPort1.Open()
        Catch
            MessageBox.Show("Error: COM port not available.", "My Application", _
                       MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
            Exit Sub
        End Try

        ' init the PIL-Box in control-on mode 
        If InitPILBox(CON) < 0 Then
            MessageBox.Show("Error: no response from PILBox.", "My Application", _
                       MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
            SerialPort1.Close()
            Exit Sub
        End If

        ButtonStop.Enabled = True
        ButtonStart.Enabled = False
        GroupBox_Ctrl.Enabled = True
        TextBox_ComPort.Enabled = False
        RadioButton_lowspeed.Enabled = False
        RadioButton_highspeed.Enabled = False

    End Sub

    Private Sub StopPILbox()
        Try
            InitPILBox(TDIS)   ' stop the PIL-Box       
        Catch
            ' in case the PILBox is disconnected before closing ILvlif...
        End Try
        Try
            SerialPort1.Close()
        Catch
            ' in case the PILBox is disconnected before closing ILvlif...
        End Try

    End Sub


    ' --------------------------------
    ' STOP button
    ' --------------------------------
    Private Sub ButtonStop_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ButtonStop.Click
        ButtonStop.Enabled = False
        ButtonStart.Enabled = True
        GroupBox_Ctrl.Enabled = False
        TextBox_ComPort.Enabled = True
        RadioButton_lowspeed.Enabled = True
        RadioButton_highspeed.Enabled = True
        StopPILbox()
    End Sub


    ' --------------------------------
    ' Form closing
    ' --------------------------------
    Private Sub Form1_FormClosing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
        StopPILbox()
    End Sub



    ' --------------------------------
    ' Command buttons
    ' --------------------------------

    ' "Media Info"
    Private Sub Button_info_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button_info.Click
        Dim addr As Integer, a As Integer
        Dim buf(256) As Byte

        addr = TextBox_addr.Text

        a = CheckMassStorage(addr, buf)
        If a < 0 Then
            Exit Sub
        End If

        ' decode information from record 0
        DisplayInfo(buf)

    End Sub


    ' "Catalog"
    Private Sub Button_catalog_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button_catalog.Click
        Dim addr As Integer, a As Integer, i As Integer
        Dim buf(256) As Byte
        Dim endcat As Boolean
        Dim catlen As Integer

        addr = TextBox_addr.Text

        a = CheckMassStorage(addr, buf)
        If a < 0 Then
            Exit Sub
        End If

        a = buf(10) : a = (a << 8) + buf(11)
        If a <> 2 Then
            DisplayStr("Warning, wrong start of catalog: " & a & CRLF)
        End If

        a = buf(18) : a = (a << 8) + buf(19)
        catlen = a
        If catlen > 80 Then
            DisplayStr("Warning, excessive length of catalog: " & catlen & CRLF)
            catlen = 80
        End If

        SeekRecord(addr, 2)  ' start of catalog

        a = ReadStatus(addr)
        If a <> 0 Then
            DisplayStr("Error " & a & CRLF)
            Exit Sub
        End If

        DisplayStr("File        Type   Len.    Date    Hour" & CRLF)

        ReadRecord(addr)

        Do

            a = ReadStatus(addr)
            If a <> 0 Then
                DisplayStr("Error " & a & CRLF)
                Exit Sub
            End If

            ' display the 8 entries/sector
            For i = 1 To 8
                a = ReadBuffer(addr, buf, 32)
                If a <> 32 Then
                    DisplayStr("Read Error " & a & CRLF)
                    Exit Sub
                End If

                ' decode catalog information
                endcat = DisplayCatalog(buf)
                If endcat Then Exit For

            Next i

            My.Application.DoEvents()

            catlen = catlen - 1 : If catlen = 0 Then endcat = True

        Loop Until endcat

    End Sub


    ' "Transfer to image file"
    Private Sub Button_transfer_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button_transfer.Click
        '        SaveFileDialog1.ShowDialog()
        MessageBox.Show("Not implemented", "ILvlif", _
                   MessageBoxButtons.OK, MessageBoxIcon.Exclamation)

    End Sub

    ' save to image file (not tested)
    Private Sub savefile(ByVal sender As System.Object, ByVal e As System.ComponentModel.CancelEventArgs) Handles SaveFileDialog1.FileOk
        Dim addr As Integer, a As Integer, i As Integer
        Dim buf(256) As Byte
        Dim imagefilename As String
        Dim mystream As Stream

        imagefilename = SaveFileDialog1.FileName
        DisplayStr("Saving LIF volume to: " & imagefilename & CRLF)

        mystream = SaveFileDialog1.OpenFile()
        addr = TextBox_addr.Text

        My.Application.DoEvents()

        a = CheckMassStorage(addr, buf)
        If a >= 0 Then
            mystream.Write(buf, 0, 256)
            ' read sectors until an error happens... or 512 sectors (HP82181 cassette drive) have been read
            For i = 1 To 511
                DisplayStr("Sector " & i)
                ReadRecord(addr)
                a = ReadStatus(addr)
                If a <> 0 Then Exit For
                a = ReadBuffer(addr, buf, 256)
                mystream.Write(buf, 0, 256)
                DisplayStr(CRLF)
                My.Application.DoEvents()
            Next i
        End If
        mystream.Close()

    End Sub




    ' check the mass storage unit, read sector 0, check if LIF volume
    Function CheckMassStorage(ByVal addr As Integer, ByRef buf() As Byte) As Integer
        Dim a As Integer

        a = ReadAID(addr)
        If a < 0 Then
            DisplayStr("No response to Acc. ID" & CRLF)
            Return -1
            Exit Function
        ElseIf (a <> 16) Then
            DisplayStr("Not a mass storage unit" & CRLF)
            Return -1
            Exit Function
        End If

        a = ReadStatus(addr)
        If a < 0 Then
            DisplayStr("No response to Status" & CRLF)
            Return -1
            Exit Function
        End If
        If a = 23 Then
            ' new disc, read again
            a = ReadStatus(addr)
        End If
        If a = 20 Then
            ' no disc
            DisplayStr("No disc" & CRLF)
            Return -1
            Exit Function
        End If

        SeekRecord(addr, 0)

        a = ReadStatus(addr)
        If a <> 0 Then
            DisplayStr("Error " & a & CRLF)
            Return -1
            Exit Function
        End If

        ReadRecord(addr)

        a = ReadStatus(addr)
        If a <> 0 Then
            DisplayStr("Error " & a & CRLF)
            Return -1
            Exit Function
        End If

        a = ReadBuffer(addr, buf, 256)
        If a <> 256 Then
            DisplayStr("Read Error " & a & CRLF)
            Return -1
            Exit Function
        End If


        a = buf(0) : a = (a << 8) + buf(1)
        If (a <> &H8000) Then
            DisplayStr("Unknown disc" & CRLF)
            Return -1
            Exit Function
        End If

        Return 0
    End Function



    ' display the information from sector 0
    Sub DisplayInfo(ByVal buf() As Byte)
        Dim label As String, stamp As String
        Dim a As Integer

        a = buf(0) : a = (a << 8) + buf(1)
        If (a <> &H8000) Then
            DisplayStr("Unknown disc" & CRLF)
        Else
            DisplayStr("LIF disc" & CRLF)
        End If
        If buf(2) <> 0 Then
            label = Chr(buf(2)) & Chr(buf(3)) & Chr(buf(4)) & Chr(buf(5)) & Chr(buf(6)) & Chr(buf(7))
        Else
            label = ""
        End If
        DisplayStr("Label : " & label & CRLF)

        a = buf(10) : a = (a << 8) + buf(11)
        DisplayStr("Catalog start  : " & a & CRLF)

        a = buf(18) : a = (a << 8) + buf(19)
        DisplayStr("Catalog length  : " & a & CRLF)

        a = buf(20) : a = (a << 8) + buf(21)
        DisplayStr("Version  : " & a & CRLF)

        If (a <> 0) Then

            a = buf(26) : a = (a << 8) + buf(27)
            DisplayStr("Tracks/surface : " & a & CRLF)

            a = buf(30) : a = (a << 8) + buf(31)
            DisplayStr("Surface number : " & a & CRLF)

            a = buf(34) : a = (a << 8) + buf(35)
            DisplayStr("Sectors/track : " & a & CRLF)

            a = buf(36) : stamp = Chr(48 + (a >> 4)) & Chr(48 + (a And 15)) & "/"
            a = buf(37) : stamp = stamp & Chr(48 + (a >> 4)) & Chr(48 + (a And 15)) & "/"
            a = buf(38) : stamp = stamp & Chr(48 + (a >> 4)) & Chr(48 + (a And 15)) & " "
            a = buf(39) : stamp = stamp & Chr(48 + (a >> 4)) & Chr(48 + (a And 15)) & ":"
            a = buf(40) : stamp = stamp & Chr(48 + (a >> 4)) & Chr(48 + (a And 15)) & ":"
            a = buf(41) : stamp = stamp & Chr(48 + (a >> 4)) & Chr(48 + (a And 15))
            DisplayStr("Initialized on : " & stamp & CRLF)

        End If

    End Sub

    ' display a catalog entry
    Function DisplayCatalog(ByVal record() As Byte) As Boolean
        Dim name As String
        Dim stamp As String
        Dim slg As String
        Dim a As Integer
        Dim lg As Integer
        Dim type As Integer


        If record(0) = 0 Then
            Return True
            Exit Function
        End If
        name = Chr(record(0)) & Chr(record(1)) & Chr(record(2)) & Chr(record(3)) & Chr(record(4))
        name = name & Chr(record(5)) & Chr(record(6)) & Chr(record(7)) & Chr(record(8)) & Chr(record(9))

        type = record(10) : type = (type << 8) + record(11)
        If type > 32767 Then type = type - 65536 ' for signed 16-bit number
        If type = -1 Then
            Return True
            Exit Function
        End If

        lg = record(18) : lg = (lg << 8) + record(19)
        If (lg < 256) Then
            slg = lg * 256
        Else
            slg = (lg \ 4) & "K"
        End If
        While Len(slg) < 5
            slg = " " & slg
        End While

        a = record(20) : stamp = Chr(48 + (a >> 4)) & Chr(48 + (a And 15)) & "/"
        a = record(21) : stamp = stamp & Chr(48 + (a >> 4)) & Chr(48 + (a And 15)) & "/"
        a = record(22) : stamp = stamp & Chr(48 + (a >> 4)) & Chr(48 + (a And 15)) & " "
        a = record(23) : stamp = stamp & Chr(48 + (a >> 4)) & Chr(48 + (a And 15)) & ":"
        a = record(24) : stamp = stamp & Chr(48 + (a >> 4)) & Chr(48 + (a And 15)) & ":"
        a = record(25) : stamp = stamp & Chr(48 + (a >> 4)) & Chr(48 + (a And 15))

        DisplayStr(name & " ")

        Select Case type
            Case 0 : DisplayStr("(NUL)")
            Case 1 : DisplayStr("ASCII")
            Case -8184 : DisplayStr("BIN8x")
            Case -8176 : DisplayStr("DTA8x")
            Case -8160 : DisplayStr("BAS8x")
            Case -8128 : DisplayStr("ALL41")
            Case -8112 : DisplayStr("KEY41")
            Case -8110 : DisplayStr("TXT75")
            Case -8109 : DisplayStr("APP75")
            Case -8080 : DisplayStr("X-M41")
            Case -8064 : DisplayStr("PGM41")
            Case -8056 : DisplayStr("BAS75")
            Case -8055 : DisplayStr("LEX75")
            Case -8054 : DisplayStr("WKS75")
            Case -7984 : DisplayStr("SDATA")
            Case -7951, -7952 : DisplayStr("DTA71")
            Case -7676 : DisplayStr("BIN71")
            Case -7669, -7670, -7671, -7672 : DisplayStr("LEX71")
            Case -7667, -7668 : DisplayStr("KEY71")
            Case -7656 : DisplayStr("FTH71")
            Case -7657, -7658, -7659, -7660 : DisplayStr("BAS71")
            Case -7652 : DisplayStr("ROM71")
            Case -5822 : DisplayStr("SYS9k")
            Case -5813 : DisplayStr("HP-UX")
            Case -5808 : DisplayStr("BAS9k")
            Case -5791 : DisplayStr("BDA9k")
            Case -5775 : DisplayStr("BIN9k")
            Case -5622 : DisplayStr("DTA9k")
            Case -5582 : DisplayStr("COD9k")
            Case -5570 : DisplayStr("TXT9k")
            Case Else : DisplayStr(type)
        End Select
        DisplayStr("  " & slg)
        DisplayStr("  " & stamp & CRLF)

        Return False  ' not the end of the catalog

    End Function



    ' ------------------------------------------------------------
    ' HP-IL mass storage primitives
    ' ------------------------------------------------------------

    ' read the AID
    Function ReadAID(ByVal addr As Integer) As Integer
        Dim b(2) As Byte

        If GetBytes(addr, SAI, b, 1) > 0 Then
            Return b(0)
        Else
            Return -1
        End If

    End Function

    ' read status, wait for drive ready
    Function ReadStatus(ByVal addr As Integer) As Integer
        Dim b(2) As Byte
        Dim stat As Integer

        Do
            If GetBytes(addr, SST, b, 1) > 0 Then
                stat = b(0)
            Else
                stat = -1
            End If
        Loop While stat >= 32  ' while drive is busy

        Return stat

    End Function

    ' seek record: DDL 4 data_high data_low
    Function SeekRecord(ByVal addr As Integer, ByVal n As Integer) As Integer
        StartTrans()
        SendFrame(LAD + addr)
        SendFrame(DDL + 4)  ' DDL 4
        SendFrame(n >> 8)
        SendFrame(n And 255)
        EndTrans()

    End Function


    ' read current record: DDT 2
    Function ReadRecord(ByVal addr As Integer) As Integer
        StartTrans()
        SendFrame(TAD + addr)
        SendFrame(DDT + 2)  ' DDT 2
        EndTrans()
    End Function


    Function ReadBuffer(ByVal addr As Integer, ByVal buf() As Byte, ByVal n As Integer) As Integer
        Return GetBytes(addr, SDA, buf, n)
    End Function


    ' ------------------------------------------------------------
    ' HP-IL primitives
    ' ------------------------------------------------------------

    ' HP-IL frame definition
    ' command frames
    Const _GET = &H408 '  (GET is a VB reserved word)
    Const DCL = &H414
    Const LAD = &H420
    Const UNL = &H43F
    Const TAD = &H440
    Const UNT = &H45F
    Const IFC = &H490
    Const REN = &H492
    Const NRE = &H493
    Const AAU = &H49A
    Const DDL = &H4A0
    Const DDT = &H4C0
    ' ready frames
    Const ETO = &H540
    Const ETE = &H541
    Const NRD = &H542
    Const SDA = &H560
    Const SST = &H561
    Const SDI = &H562
    Const SAI = &H563
    Const AAD = &H580


    ' starts a HPIL transaction (IFC, AAU, AAD 1)
    Private Function StartTrans()
        If SendFrame(IFC) <> IFC Then GoTo err
        If SendFrame(AAU) <> AAU Then GoTo err
        If SendFrame(AAD + 1) < 0 Then GoTo err
        Return 0
err:    Return -1
    End Function

    ' ends a HPIL transaction (UNL, UNT)
    Private Sub EndTrans()
        SendFrame(UNL)
        SendFrame(UNT)
    End Sub


    ' send an Universal Command Group frame
    Private Sub SendUCG(ByVal frame As Integer)
        If StartTrans() < 0 Then GoTo err
        SendFrame(frame)
        Exit Sub
err:
        DisplayStr("TRANSMIT ERR" & CRLF)
    End Sub

    ' send an Addressed Command Group frame
    Private Sub SendACG(ByVal addr As Integer, ByVal frame As Integer)
        If StartTrans() < 0 Then GoTo err
        SendFrame(LAD + addr) ' address device as listener
        SendFrame(frame)
        EndTrans()
        Exit Sub
err:
        DisplayStr("TRANSMIT ERR" & CRLF)
    End Sub


    ' send a data string to device at addr 
    Private Sub SendData(ByVal addr As Integer, ByVal s As String)
        Dim i As Integer

        If StartTrans() < 0 Then GoTo err
        SendFrame(LAD + addr) ' address device as listener
        For i = 1 To Len(s)
            SendFrame(Asc(Mid(s, i, i)))
        Next i
        EndTrans()
        Exit Sub
err:
        DisplayStr("TRANSMIT ERR" & CRLF)
    End Sub


    ' get a data string from device at addr n with sot frame SDA or SDI
    ' end the transmission if ETO/ETE received,
    ' interrupts the transfer if LF received,  END byte received or 80 bytes received
    Private Sub GetData(ByVal addr As Integer, ByVal sot As Integer, ByRef s As String)
        Dim frame As Integer
        Dim i As Integer

        If StartTrans() < 0 Then GoTo err
        SendFrame(TAD + addr) ' address device as talker
        s = "" : i = 0
        frame = SendFrame(sot) ' send start of transmission frame (SDI or SDA)
        ' if sot frame returns, it means the addressed device didn't answer (this is not an error)
        If (frame <> sot) Then
            Do
                If (frame = ETO) Then Exit Do ' end of transmission ok
                If (frame = ETE) Then Exit Do ' end of transmission error
                If (frame < &H400) Then
                    ' data or end frames
                    s = s & Chr(frame And &HFF) ' append to string
                    i = i + 1
                    If ((frame And &H4FF) = 10) Or ((frame And &H600) = &H200) Or (i > 80) Then
                        ' LF received or END byte received or 80 bytes received,
                        ' insert a  Not Ready for Data  frame to interrupt transfer
                        If SendFrame(NRD) <> NRD Then GoTo err
                    End If
                End If
                frame = SendFrame(frame) ' retransmit frame and get next one
            Loop
        End If
        EndTrans()
        If (frame = ETE) Then GoTo err
        Exit Sub
err:
        DisplayStr("TRANSMIT ERR" & CRLF)
    End Sub


    ' get binary data from device at addr with sot frame SDA, SAI or SST
    ' end the transmission if ETO/ETE received,
    ' interrupts the transfer after n bytes received
    Private Function GetBytes(ByVal addr As Integer, ByVal sot As Integer, ByRef s() As Byte, ByVal n As Integer) As Integer
        Dim frame As Integer
        Dim i As Integer

        If StartTrans() < 0 Then GoTo err
        SendFrame(TAD + addr) ' address device n as talker
        i = 0
        frame = SendFrame(sot) ' send start of transmission frame (SAI, SST or SDA)
        ' if sot frame returns, it means the addressed device didn't answer (this is not an error)
        If (frame <> sot) Then
            Do
                If (frame = ETO) Then Exit Do ' end of transmission ok
                If (frame = ETE) Then Exit Do ' end of transmission error
                If (frame < &H400) Then
                    ' data or end frames
                    s(i) = frame And &HFF
                    i = i + 1
                    If (i >= n) Then
                        ' insert a  Not Ready for Data  frame to interrupt transfer
                        If SendFrame(NRD) <> NRD Then GoTo err
                    End If
                End If
                frame = SendFrame(frame) ' retransmit frame and get next one
            Loop
        End If
        EndTrans()
        If (frame = ETE) Then GoTo err
        Return i
        Exit Function
err:
        DisplayStr("TRANSMIT ERR" & CRLF)
        Return -1
    End Function




    ' ------------------------------------------------------------
    ' PILBox functions
    ' ------------------------------------------------------------

    ' --------------------------------
    ' InitPILBox()
    ' init the PIL Box 
    ' --------------------------------
    Private Function InitPILBox(ByVal cmd As Integer) As Integer

        lasth = 0
        If SendFrame(cmd) <> cmd Then
            Return -1  ' error
        Else
            Return 0
        End If

    End Function


    ' --------------------------------
    ' send a frame to the PILBox, wait for the returned frame
    ' frame<0 means no frame returns (timeout)
    ' --------------------------------
    Private Function SendFrame(ByVal frame As Integer) As Integer
        Dim lbyt As Integer, hbyt As Integer  ' low/high bytes to serial
        Dim buf(2) As Byte, byt As Byte
        Dim n As Integer
        Dim s As String

        ' build the low and high parts
        If Not USE8BITS Then
            ' use 7-bit characters for maximum compatibility
            hbyt = ((frame >> 6) And &H1F) Or &H20
            lbyt = (frame And &H3F) Or &H40
        Else
            ' 8-bit format for optimum speed
            hbyt = ((frame >> 6) And &H1E) Or &H20
            lbyt = (frame And &H7F) Or &H80
        End If
        If (hbyt <> lasth) Then
            ' send high part if different from last one
            lasth = hbyt
            buf(1) = hbyt : buf(2) = lbyt : n = 2
        Else
            ' otherwise send only low part
            buf(1) = lbyt : n = 1
        End If
        SerialPort1.DiscardInBuffer()
        SerialPort1.Write(buf, 1, n)

        ' scope - can be inserted before or after frame received
        s = ILMnemo(frame)
        DisplayMnemo(s)

        ' wait for returned frame
        ' timeouts: 10s for DOE/RDY, 10s for CMD, 1s for IDY 
        If frame < &H400 Then
            Timer1.Interval = 10000 ' DOE
        ElseIf (frame And &H7F4) = &H494 Then
            Timer1.Interval = 500  ' special PILBox frames 494-497 and 49c-49f
        ElseIf frame < &H500 Then
            Timer1.Interval = 10000  ' CMD
        ElseIf frame < &H600 Then
            Timer1.Interval = 10000  ' RDY
        Else
            Timer1.Interval = 1000  ' IDY
        End If
        frame = -1
        frame_timeout = False
        Timer1.Enabled = True

        Do

            Try
                byt = SerialPort1.ReadByte()
            Catch
                byt = 0
            End Try
            If byt <> 0 Then
                If ((byt And &HC0) = 0) Then
                    If (byt And &H20) Then
                        ' high byte, save it
                        lasth = byt
                        SerialPort1.Write(Chr(13))   ' acknowledge
                    End If
                Else
                    ' low byte, build frame according to format 
                    If (byt And &H80) Then
                        frame = ((lasth And &H1E) << 6) + (byt And &H7F)
                    Else
                        frame = ((lasth And &H1F) << 6) + (byt And &H3F)
                    End If
                End If
            End If

            My.Application.DoEvents()

        Loop Until frame_timeout Or (frame >= 0)

        Timer1.Enabled = False
        Return frame

    End Function


    ' handler for receive timeout
    Private Sub Timer1_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Timer1.Tick
        frame_timeout = True
    End Sub


End Class


